import os
import struct
import math
import bpy
import mathutils
import bmesh
from bpy_extras.io_utils import ImportHelper



def skipPad(f, blockSize, force):
    if force or (f.tell() % blockSize != 0):
        pad = blockSize - (f.tell() % blockSize)
        f.read(pad)
        
def readString(f):
    out = ""
    while True:
        c = f.read(1)
        if len(c) == 0: break
        if c == b"\x00":
            if len(out) == 0: continue
            else: break
        out = out + c.decode("ascii")
    return out

def readPointer(f):
    return struct.unpack( "Q", f.read(8) )[0]

def readInt(f):
    return struct.unpack( "I", f.read(4) )[0]

def readShort(f):
    return struct.unpack( "H", f.read(2) )[0]

def readFloat(f):
    return struct.unpack( "f", f.read(4) )[0]

def readByte(f):
    return struct.unpack( "B", f.read(1) )[0]

def readVector3(f):
    vec = struct.unpack( "fff", f.read(12) )
    return [ vec[0], vec[1], vec[2] ]

def writeInt(f, n):
    f.write( struct.pack("I", n) )

def writeShort(f, n):
    f.write( struct.pack("H", n) )

def writeFloat(f, n):
    f.write( struct.pack("f", n) )

def writeByte(f, n):
    f.write( struct.pack("B", n) )

def writeVector3(f, n):
    f.write( struct.pack("fff", n[0], n[1], n[2]) )


class CollisionPrimitiveInfo:
    def write(self, w):
        writeInt(w, self.bone_index)
        writeShort(w, self.is_capsule)
        writeShort(w, self.do_scale)
        writeFloat(w, self.radius)
        writeFloat(w, self.height)
        writeVector3(w, self.pos)
        writeVector3(w, self.axis)
        writeVector3(w, self.local_space_pos)

    def __init__(self, f):
        self.bone_index = readInt(f)
        self.is_capsule = readShort(f)
        self.do_scale = readShort(f)
        self.radius = readFloat(f)
        self.height = readFloat(f)
        self.pos = readVector3(f)
        self.axis = readVector3(f)
        self.local_space_pos = readVector3(f)

class RopeInfo:
    def writeIndices(self, w, exp):
        for i in range(self.num_nodes):
            writeInt(w, self.node_indecies[i])
        for i in range(self.num_links):
            writeInt(w, self.link_indecies[i])

    def write(self, w, exp):
        writeFloat(w, exp)
        writeInt(w, self.num_nodes)
        writeInt(w, self.num_links)
        w.write(self.pad1)

    def readIndices(self, f):
        self.node_indecies = [ readInt(f) for i in range(self.num_nodes) ]
        self.link_indecies = [ readInt(f) for i in range(self.num_links) ]

    def __init__(self, f):
        self.length = readFloat(f)
        self.num_nodes = readInt(f)
        self.num_links = readInt(f)
        self.pad1 = f.read(8)
        self.node_indecies = list()
        self.link_indecies = list()

class SimNodeLinkInfo:
    def write(self, w, exp):
        writeShort(w, self.node_index_1)
        writeShort(w, self.node_index_2)
        writeInt(w, self.collide)
        writeFloat(w, exp["length"])
        writeFloat(w, exp["stretch_len"])
        writeFloat(w, exp["twist"])
        writeFloat(w, exp["spring"])
        writeFloat(w, exp["damp"])

    def __init__(self, f):
        self.node_index_1 = readShort(f)
        self.node_index_2 = readShort(f)
        self.collide = readInt(f) # bit list of colliders that this linke collides with
        self.length = readFloat(f)
        self.stretch_len = readFloat(f)
        self.twist = readFloat(f)
        self.spring = readFloat(f)
        self.damp = readFloat(f)

class SimNodeInfo:
    def write(self, w, exp):
        writeByte(w, self.bone_index)
        writeByte(w, self.parent_node_index)
        writeByte(w, self.gravity_link)
        writeByte(w, self.anchor)
        writeInt(w, self.collide)
        writeFloat(w, self.wind_multiplier)
        writeVector3(w, self.pos)
        writeVector3(w, exp)

    def __init__(self, f):
        self.bone_index = readByte(f)
        self.parent_node_index = readByte(f) # rotation is applied to the parent bone for each particle
        self.gravity_link = readByte(f)
        self.anchor = readByte(f)
        self.collide = readInt(f) # binary list of colliders that this node collides with.
        self.wind_multiplier = readFloat(f)
        self.pos = readVector3(f)
        self.localPos = readVector3(f)

class SimFile:
    def write(self, exp):
        w = open(self.filePath + ".out", "wb")
        w.write(self.header)
        writeInt(w, exp["p"]["numPasses"])
        writeFloat(w, exp["p"]["air_resistance"])
        writeFloat(w, exp["p"]["wind_multiplier"])
        writeFloat(w, exp["p"]["wind_const"])
        writeFloat(w, exp["p"]["gravity_multiplier"])
        writeFloat(w, exp["p"]["object_velocity_inheritance"])
        writeFloat(w, exp["p"]["object_position_inheritance"])
        writeFloat(w, exp["p"]["object_rotation_inheritance"])
        writeInt(w, exp["p"]["wind_type"])
        writeInt(w, self.num_nodes)
        writeInt(w, self.num_anchor_nodes)
        writeInt(w, self.num_node_links)
        writeInt(w, self.num_ropes)
        writeInt(w, self.num_colliders)
        w.write(self.pad2)
        for i in range(self.num_nodes):
            self.node[i].write(w, exp["n"][i])
        for i in range(self.num_node_links):
            self.link[i].write(w, exp["l"][i])
        for i in range(self.num_ropes):
            self.rope[i].write(w, exp["r"][i])
        for i in range(self.num_ropes):
            self.rope[i].writeIndices(w, exp["r"][i])
        for p in self.prim:
            p.write(w)
        w.write(self.pad3)
        w.close()

    def __init__(self, fp):
        self.filePath = fp
        f = open(fp, "rb")
        self.header = f.read(36)
        self.numPasses = readInt(f)
        self.air_resistance = readFloat(f)
        self.wind_multiplier = readFloat(f)
        self.wind_const = readFloat(f)  
        self.gravity_multiplier = readFloat(f)
        self.object_velocity_inheritance = readFloat(f)
        self.object_position_inheritance = readFloat(f)
        self.object_rotation_inheritance = readFloat(f)
        self.wind_type = readInt(f)
        self.num_nodes = readInt(f)
        self.num_anchor_nodes = readInt(f)
        self.num_node_links = readInt(f)
        self.num_ropes = readInt(f)
        self.num_colliders = readInt(f)
        self.pad2 = f.read(16)
        self.node = [ SimNodeInfo(f) for i in range(self.num_nodes) ]
        self.link = [ SimNodeLinkInfo(f) for i in range(self.num_node_links) ]
        self.rope = [ RopeInfo(f) for i in range(self.num_ropes) ]
        for i in range(self.num_ropes):
            self.rope[i].readIndices(f)
        self.prim = [ CollisionPrimitiveInfo(f) for i in range(self.num_colliders) ]
        self.pad3 = f.read()
        f.close()


def getDistance(v1, v2):
    return math.sqrt( ( (v2[0] - v1[0]) * (v2[0] - v1[0]) ) + 
                        ( (v2[1] - v1[1]) * (v2[1] - v1[1]) ) + 
                        ( (v2[2] - v1[2]) * (v2[2] - v1[2]) ) )

def buildRepresentation(sf, fileName):
    # Parent object to transform the other objects
    parM = bpy.data.meshes.new(name = "Mesh")
    par = bpy.data.objects.new(name = (fileName + "-Sim"), object_data = parM)
    bpy.context.scene.collection.objects.link(par)
    bpy.context.view_layer.objects.active = par
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.primitive_uv_sphere_add(radius = 0.01, location = (0.0, 0.0, 0.0) )
    bpy.ops.object.mode_set(mode="OBJECT")
    bpy.ops.object.select_all(action="DESELECT")

    linksArm = bpy.data.armatures.new(name = "Armature")
    linksObj = bpy.data.objects.new(name = ( fileName + "-Links" ), object_data = linksArm)
    bpy.context.scene.collection.objects.link(linksObj)
    linksObj["numPasses"] = sf.numPasses
    linksObj["air_resistance"] = sf.air_resistance
    linksObj["wind_multiplier"] = sf.wind_multiplier
    linksObj["wind_const"] = sf.wind_const
    linksObj["gravity_multiplier"] = sf.gravity_multiplier
    linksObj["object_velocity_inheritance"] = sf.object_velocity_inheritance
    linksObj["object_position_inheritance"] = sf.object_position_inheritance
    linksObj["object_rotation_inheritance"] = sf.object_rotation_inheritance
    linksObj["wind_type"] = sf.wind_type
    bpy.context.view_layer.objects.active = linksObj
    linksObj.select_set(True)
    bpy.ops.object.mode_set(mode="EDIT")
    nodeList = [None] * sf.num_nodes
    for i in range(sf.num_ropes):
        linkIndices = sf.rope[i].link_indecies
        parentLink = None
        for l in linkIndices:
            link = sf.link[l]
            bone = linksArm.edit_bones.new( "{:02}-Link".format(l) )
            n1 = sf.node[ link.node_index_1 ].localPos
            n2 = sf.node[ link.node_index_2 ].localPos
            bone.head = n1
            bone.tail = n2
            nodeList[ link.node_index_1 ] = bone.head
            nodeList[ link.node_index_2 ] = bone.tail
            bone["stretch_len"] = link.stretch_len
            bone["twist"] = link.twist
            bone["spring"] = link.spring
            bone["damp"] = link.damp
            if parentLink != None: 
                bone.parent = parentLink
                bone.use_connect = True
            parentLink = bone
    for i in range(sf.num_nodes):
        if nodeList[i] == None:
            bone = linksArm.edit_bones.new( "{:02}-Node".format(i) )
            bone.head = sf.node[i].localPos
            bone.tail = ( sf.node[i].localPos[0], sf.node[i].localPos[1], (sf.node[i].localPos[2] + 0.01) )
            nodeList[i] = bone.head
    bpy.ops.object.mode_set(mode="OBJECT")
    bpy.ops.object.select_all(action="DESELECT")
    linksObj.parent = par

    # Collision primitives
    for i in range(sf.num_colliders):
        prim = sf.prim[i]
        rad = prim.radius
        dep = prim.height
        pos = prim.local_space_pos
        v0 = mathutils.Vector((0.0, 0.0, 1.0))
        v1 = mathutils.Vector(prim.axis)
        rot = v1.rotation_difference(v0).to_euler()
        colM = bpy.data.meshes.new(name = "Mesh")
        col = bpy.data.objects.new(name = ( fileName + "-{:02}".format(i) + "-Collider" ), object_data = colM)
        bpy.context.scene.collection.objects.link(col)
        bpy.context.view_layer.objects.active = col
        bpy.ops.object.mode_set(mode="EDIT")
        bpy.ops.mesh.primitive_cylinder_add(radius = rad, depth = dep, location = (0.0, 0.0, 0.0) )
        bm = bmesh.from_edit_mesh(colM)
        for v in bm.verts:
            v.co[2] += dep/2.0
        bmesh.update_edit_mesh(colM)
        bpy.ops.object.mode_set(mode="OBJECT")
        col.location = pos
        col.rotation_euler = [ (rot[0] * -1.0), (rot[1] * -1.0), (rot[2] * 1.0) ]
        col.parent = par
        bpy.ops.object.select_all(action="DESELECT")
    
    # Transform the parent object to fit the body template
    bpy.context.view_layer.objects.active = par
    par.scale = (-0.393701, 0.393701, 0.393701)
    par.rotation_euler = (1.5707963, 0.0, 0.0)
    bpy.ops.object.select_all(action="DESELECT")

def getExportData(obj, sim):
    nodes = [None] * sim.num_nodes
    links = [ dict() for i in range(sim.num_node_links) ]
    ropes = [0.0] * sim.num_ropes
    param = dict()
    baseParams = obj.keys()
    for k in baseParams:
        param[k] = obj[k]
    for i in range(sim.num_ropes):
        linkIndices = sim.rope[i].link_indecies
        ropeLen = 0.0
        for l in linkIndices:
            link = sim.link[l]
            boneName = "{:02}-Link".format(l)
            linkBone = obj.data.bones[boneName]
            nodes[ link.node_index_1 ] = linkBone.head_local
            nodes[ link.node_index_2 ] = linkBone.tail_local
            linkLen = getDistance(linkBone.head_local, linkBone.tail_local)
            ropeLen += linkLen
            links[l]["length"] = linkLen
            linkParams = linkBone.keys()
            for k in linkParams:
                links[l][k] = linkBone[k]
        ropes[i] = ropeLen
    #print(nodes)
    #print(links)
    #print(ropes)
    #print(param)
    for i in range(sim.num_nodes):
        if nodes[i] == None:
            boneName = "{:02}-Node".format(i)
            nodeBone = obj.data.bones[boneName]
            nodes[i] = nodeBone.head
    out = dict()
    out["n"] = nodes
    out["l"] = links
    out["r"] = ropes
    out["p"] = param
    return out


class ImportSimFile(bpy.types.Operator, ImportHelper):
    from bpy.props import ( StringProperty, BoolProperty )
    """Saints Row 3 mesh importer"""        # Use this as a tooltip for menu items and buttons.
    bl_idname = "import_scene.sr3sim"       # Unique identifier for buttons and menu items to reference.
    bl_label = "Import sim"                 # Display name in the interface.
    bl_options = {'UNDO'}                   # Enable undo for the operator.
    filename_ext = ".sim_pc"
    filter_glob: StringProperty(default="*.sim_pc", options={'HIDDEN'})
    exportFile: BoolProperty( 
        name = "Export", 
        description = "Export file with the same name and .out extension.", 
        default = False, 
        )
    
    def execute(self, context):             # execute() is called when running the operator.
        sim = SimFile(self.filepath)
        fileName = self.filepath.split("\\")[-1].split(".")[0]
        if self.exportFile:
            obj = bpy.context.active_object
            exp = getExportData(obj, sim)
            sim.write(exp)
        else: buildRepresentation(sim, fileName)
        return {"FINISHED"}

bpy.utils.register_class(ImportSimFile)
bpy.ops.import_scene.sr3sim("INVOKE_DEFAULT")






